<?php
    /**
    * This file is the basic core engine for templates
    * in DynPG.
    *
	* The class TTemplate implements the basic template logic and
    * enables the user to parse a string/file.
    *
    *
    * @version 2.2
	* @package DynPG PlugIn Engine
    * @author Daniel Schliebner
    *
    * (c) 2010 Daniel Schliebner
    */
    class TTemplate {
		// array $rows
    	// Array that contains placeholders and their data, which is replaced
		// inside the templates.
		var $rows;

        // boolean $autoRender
    	// If autoRender is boolean true, pseudo if- and php-statements
		// will be rendered directly, otherwise they are kept inside
		// to be replaced later.
		var $autoRender;

		// array $pVarStack
		// Array that contains variables (and their data), declared inside
		// a template using the pseudocode statements {$#SET #} and {$#GET #}.
		var $pVarStack;

        // boolean $templateDebug
        // If templateDebug is boolean true, template var $TEMPLATE_DEBUG
        // will be filled with debugging information.
        var $templateDebug;


		/**
		 * TTemplate::TTemplate()
		 *
		 * Constructor of this class.
		 */
		function TTemplate() {
            // constructor //

            $this->rows = Array();
            $this->autoRender = true;
            $this->pVarStack = Array();
            $this->templateDebug = false;

            register_shutdown_function(array(&$this, 'destroy_TTemplate'));
        }


		/**
         * TTemplate::refresh()
         *
		 * Sets the placeholder-property $rows to an empty array.
         *
         * @return
         */
        function refresh() {
            $this->rows = Array();
            $this->pVarStack = Array();
        }


		/**
         * TTemplate::autoRenderOn()
         *
         * Enables automatic rendering (see description of variable $autoRender)
         *
         * @return
         */
        function autoRenderOn() {
        	$this->autoRender = true;
        }


		/**
         * TTemplate::autoRenderOn()
         *
         * Disables automatic rendering (see description of variable $autoRender)
         *
         * @return
         */
		function autoRenderOff() {
        	$this->autoRender = false;
        }


		/**
         * TTemplate::addVar()
         *
         * Add a global placeholder variable. Global variables are replaced, when
         * rendering a file/string, without specifing a additional array $subRows
         * in the parameter-list of a rendering method (fillTemplate or render).
         *
         * @param string $id
         * @param string $var
         * @param boolean $overwrite
         * @param boolean $scriptVarCallback
         * @return
         */
        function addVar($id = '', $var = '', $overwrite = true, $scriptVarCallback = null) {
            if ( is_callable($scriptVarCallback) ) {
                $var = $scriptVarCallback($var);
            }

            if ( !$overwrite && array_key_exists($id, $this->rows) ) {
                return;
            }

            $this->rows[$id] = $var;
        }


        /**
         * TTemplate::getTemplate()
         *
         * Reads a template file without parsing it.
         *
         * @param string $template
         * @return string
         */
        function getTemplate($template = null) {
            if (@file_exists($template)) {
                clearstatcache();
                return file_get_contents($template);
            }
            return $template;
        }


		/**
         * TTemplate::getSubTemplate()
         *
         * Reads a subTemplate inside a string/file with the name $name.
         * Set $file to false if you want to parse a string instead.
         *
         * @param string $name
         * @param mixed $template
         * @param boolean $file
         * @return string
         */
        function getSubTemplate($name = '', $template = null, $file = true) {

            // get template data from string or file

				$_ = $template;

	            if ( $file ) {
                    $_ = $this->getTemplate($template);
	            }


			// match the subtemplate area

	            if ( preg_match('/<!--[[:space:]]*RepeatedListBegin[[:space:]]*name[[:space:]]*=[[:space:]]*\"{0,1}' . (($name != '')? $name: '[^\"[:space:]]+') . '\"{0,1}[[:space:]]*-->/', $_, $regExBegin) ) {
	                $_after_ = $after = substr($_, $_C1 = (strpos($_, $regExBegin[0]) + strlen($regExBegin[0])), strlen($_) - $_C1);

	                $_d_ = 0;

	                while ( !(isset($end)) || ( isset($end) && !$end ) ) {

	                    $_d_++;

	                    preg_match('/<!--[[:space:]]*RepeatedListFinish[^-]*-->/', $_after_, $regExEnd);

	                    $__ = substr($_after_, 0, strpos($_after_, $regExEnd[0]));

	                    $end = !preg_match('/<!--[[:space:]]*RepeatedListBegin[[:space:]]*name[[:space:]]*=[[:space:]]*\"{0,1}[^\"[:space:]]+\"{0,1}[[:space:]]*-->/', $__, $regExOff);

	                    if ( !$end ) {
	                        $_after_ = str_replace($regExOff[0], str_repeat(' ', strlen($regExOff[0])), $_after_);
	                        $_after_ = str_replace($regExEnd[0], str_repeat(' ', strlen($regExEnd[0])), substr($_after_, 0, $_Cx = (strpos($_after_, $regExEnd[0]) + strlen($regExEnd[0])) )) . substr($_after_, $_Cx, strlen($_after_) - $_Cx);
	                    }

	                    if ( $end ) {
	                        $_ = substr($after, 0, strpos($_after_, $regExEnd[0]));
	                    }

	                    if ($_d_ > 999) {
							echo '<b>DoS::getSubTemplate()</b>';
							break;
						}
	                }

	                return $_;
	            }

			return false;
        }


		/**
         * TTemplate::replaceSubTemplate()
         *
         * Replaces a subTemplate inside a string/file with the name $name.
         * Set $file to false if you want to parse a string instead.
         *
         * @param string $name
         * @param string $replace
         * @param mixed $template
         * @param boolean $file
         * @return string
         */
		function replaceSubTemplate($name = '', $replace = '', $template = null, $file = true) {

            // get template data from string or file

				$_ = $template;

	            if ( $file ) {
                    $_ = $this->getTemplate($template);
	            }


			// match the subtemplate area and replace it with $replace

	            if ( preg_match('/<!--[[:space:]]*RepeatedListBegin[[:space:]]*name[[:space:]]*=[[:space:]]*\"{0,1}' . (($name != '')? $name: '[^\"[:space:]]+') . '\"{0,1}[[:space:]]*-->/', $_, $regExBegin) ) {
	                $_after_ = $after = substr($_, $_C1 = (strpos($_, $regExBegin[0]) + strlen($regExBegin[0])), strlen($_) - $_C1);

	                $_d_ = 0;

	                while ( !(isset($end)) || ( isset($end) && !$end ) ) {

	                    $_d_++;

	                    preg_match('/<!--[[:space:]]*RepeatedListFinish[^-]*-->/', $_after_, $regExEnd);

	                    $__ = substr($_after_, 0, strpos($_after_, $regExEnd[0]));

	                    $end = !preg_match('/<!--[[:space:]]*RepeatedListBegin[[:space:]]*name[[:space:]]*=[[:space:]]*\"{0,1}[^\"[:space:]]+\"{0,1}[[:space:]]*-->/', $__, $regExOff);

	                    if ( !$end ) {
	                        $_after_ = str_replace($regExOff[0], str_repeat(' ', strlen($regExOff[0])), $_after_);
	                        $_after_ = str_replace($regExEnd[0], str_repeat(' ', strlen($regExEnd[0])), substr($_after_, 0, $_Cx = (strpos($_after_, $regExEnd[0]) + strlen($regExEnd[0])) )) . substr($_after_, $_Cx, strlen($_after_) - $_Cx);
	                    }

	                    if ( $end ) {
	                        $_ = str_replace( $regExBegin[0] . substr($after, 0, strpos($_after_, $regExEnd[0])) . $regExEnd[0], $replace, $_ );
	                    }

	                    if ( $_d_ > 999 ) {
							echo '<b>DOS::replaceSubTemplate</b>'; break;
						}
	                }
	            }

            return $_;
        }

		/**
         * TTemplate::_escapeTemplateSyntax()
         *
         * Helper method to escape {$ ... } from template variable values.
         *
         * @param array &$array (call by reference)
         * @return
         */
		function _escapeTemplateSyntax(&$array) {
            if (empty($array))
				return;
			foreach ($array as $key => &$value) {
        		$value = preg_replace('/\{\$([^#]+)\}/', '%SAVE_CBRACKETL%$\1}', $value);
                dynpg_secure_template_syntax($value);
                $value = str_replace('%SAVE_CBRACKETL%', '{', $value);
        	}
		}

		/**
         * TTemplate::_unescapeTemplateSyntax()
         *
         * Helper method to unescape {$ ... } from template variable values.
         * If $string is true, then $haystack has to be a string, otherwise
         * an array.
         *
         * @param mixed &$haystack (call by reference)
         * @param boolean $string
         * @return
         */
		function _unescapeTemplateSyntax(&$haystack, $string = false) {
			if (empty($haystack))
				return;

			if (!$string)
				foreach ($haystack as $key => $value)
	        		$haystack[$key] = str_replace(array(md5('{$'),md5('#')),array('{$','#'), $value);
	        else
	        	$haystack = str_replace(array(md5('{$'),md5('#')),array('{$','#'), $haystack);
		}

		/**
         * TTemplate::fillTemplate()
         *
         * This is the heart of the template engine. This method parses a
         * complete file or a string and replaces all placeholder variables
         * (global or out of $subRows) and also interprets the pseudo if-
         * and php-statements {$# [...] #} and {$#PHP [...] #}.
         *
         * @param mixed $template
         * @param boolean $file
         * @param array $subRows
         * @return string
         */
        function emulate_call_func($call) {
        	// check if $call still contains a template variable
                if (preg_match('/\{\$([^#\}]+)\}/', $call))
                    return false;

            // fixes eval bug
            // @see https://bugs.php.net/bug.php?id=43358
            // @see https://bugs.php.net/bug.php?id=45592
                $er = error_reporting();
                $de = ini_get('display_errors');
                ini_set('display_errors', 'On');
                error_reporting(-1);

            // eval code
                @eval('$bool = ' . $call . ';');

            // reset error reporting and display_errors
                ini_set('display_errors', $de);
                error_reporting($er);

        	if (!isset($bool))
        		$bool = false;

			return $bool;
        }
        function fillTemplate($template = null, $file = true, $subRows = null, $recursive = false) {

            // get template data from string or file

				$_ = $template;

	            if ( $file ) {
                    $_ = $this->getTemplate($template);
	            }

	            unset($template);


            // skip empty strings

	           if (empty($_))
	               return '';


			// placeholder replacements
				// escape template variable syntax
				    if (!$recursive) {
    					$this->_escapeTemplateSyntax($this->rows);
    					$this->_escapeTemplateSyntax($subRows);
    				}

                // replace local placeholders
                    if ( !is_null($subRows) ) {
                        foreach ( $subRows as $key => $value ) {
                            if(strpos($_, '{$'.$key.'}') !== false)
                                $_ = str_replace('{$' . $key . '}', str_replace("\n", "", $value), $_);
                        }
                    }

                // unescape $Main-> Vars
                    if (strpos($_, md5('{$').'Main->') !== false)
                        $_ = str_replace(md5('{$').'Main->','{$Main->', $_);

	            // replace global placeholders
					if ( is_null($subRows) || $this->autoRender) {
		                foreach ( $this->rows as $key => $value ) {
                            if(strpos($_, '{$'.$key.'}') !== false)
		                      $_ = str_replace('{$' . $key . '}', str_replace("\n", "", $value), $_);
		                }
		            }

	            // replace placeholder info variable
                    $debugString = '';
		            if($this->templateDebug){
                        $debugString  = "<fieldset><legend><strong>TTemplate debug info:</strong></legend>";
                        $debugString .= "<p><strong>global</strong>:</p>";
                        foreach ( $this->rows as $key => $value ) {
                            if(is_array($value)){
                                foreach($value as $k => $v)
                                    $debugString .= "Variable <em>{$key}{$k}</em> has value: <em>" . htmlentities($v) . "</em>. <br />";
                            }else{
                                $debugString .= "Variable <em>{$key}</em> has value: <em>" . htmlentities($value) . "</em>. <br />";
                            }
                        }
                        if ( !is_null($subRows) ) {
                            $debugString .= "<p><strong>private</strong>:</p>";
                            foreach ( $subRows as $key => $value ) {
                                if(is_array($value)){
                                    foreach($value as $k => $v)
                                        $debugString .= "Variable <em>{$key}{$k}</em> has value: <em>" . htmlentities($v) . "</em>. <br />";
                                }else{
                                    $debugString .= "Variable <em>{$key}</em> has value: <em>" . htmlentities($value) . "</em>. <br />";
                                }
                            }
                        }
                        $debugString .= "</fieldset>";
		            }
                    $_ = str_replace('{$TEMPLATE_DEBUG}', str_replace("\n", "", $debugString), $_);


            // when called by recursion, then skip after placeholder replacements
                if ($recursive)
                    return $_;


            // -----------------------------------------------------------------
			// conditional replacements
            //
            // syntax: {$# assignment #} ... {$# End #}
            //

            // init variables and cache
			$CC    = Array('start' => 0, 'end' => 0, 'valid' => false, 'error' => false);
            $CACHE = Array();

            // iterate through the template code an match conditional statements
			while ( preg_match('/\{\$#[[:space:]]([^#]+)#\}/', $_, $matches) ) {

				// _C1 := position of current statement: >{$# assignment #}
				// _C2 := end of current statement {$# assignment #>}
				$_ = substr($_, 0, $_C1 = strpos($_, $matches[0])) .
                     substr($_, $_C2 = ( $_C1 + strlen($matches[0]) ), strlen($_) - $_C2);

                // // D e b u g
				// echo $file? $template: "no file";
				// echo "<br /><font color=\"red\">#begin#</font> " . htmlentities($_) . " <font color=\"red\">#end#</font>";
				// echo "<br /><b>$matches[1]</b>";
				// echo '<hr color="red">';

                // Replace if-block if no error occured:
                // if assignment was boolean true, then replace pseudo-code only,
                // otherwise replace the whole code inside the if-block.
				if ( strtolower(trim($matches[1])) == 'end' && !$CC['error'] ) {
                    // end of block is current matching point of {$# End #}
					$CC['end'] = $_C1;

                    // if assignment was boolean true, then do n o t replace
                    // the block inside the if-statement
					if ( !$CC['valid'] ) {
                        $_ = substr($_, 0, $CC['start']) .
                             substr($_, $CC['end'], strlen($_) - $CC['end']);
                    }

                    // pop current handle from stack (cache)
					if ( count($CACHE) ) {
                        $CC = $CACHE[count($CACHE) - 1];
                        $CACHE = array_slice($CACHE, 0, count($CACHE) - 1);
                    } else {
                        $CC = Array('start' => 0, 'end' => 0, 'valid' => false, 'error' => false);
                    }

                    continue;
                }

                // Skip block if an error occured: maybe he is resolvable
                // in a later step.
				if ( strtolower(trim($matches[1])) == 'end' && $CC['error'] ) {
                	// mark current {$# End #}-Tag as read to denie parsing
					// it again
					$_ = substr($_, 0, $_C1) .
						 substr($matches[0], 0, 1) . strtoupper(uniqid('@CACHED', true)) . substr($matches[0], 1, strlen($matches[0])) .
                 		 substr($_, $_C1, strlen($_) - $_C1);

					// pop current handle from stack (cache)
					if ( count($CACHE) ) {
                        $CC = $CACHE[count($CACHE) - 1];
                        $CACHE = array_slice($CACHE, 0, count($CACHE) - 1);
                    } else {
                        $CC = Array('start' => 0, 'end' => 0, 'valid' => false, 'error' => false);
                    }

                    continue;
                }

                // Push current handle to the stack (cache)
				if ( $CC['start'] > 0 ) {
                    $CACHE[] = $CC;
                }

                // // D e b u g
				// echo '<pre>';
				// print_r($CACHE);
				// echo '</pre>';
				// echo '<hr color="red">';

                // value of current assignment: has to be boolean!
				// $bool = @call_user_func( @create_function('', 'return (' . trim($matches[1]) . ');') );
				$bool = $this->emulate_call_func(trim($matches[1]));

                // set values for current handle:
                // 'valid' := value of current assignment
                // 'error' := true if assignment is not boolean, otherwise false
				$CC['valid'] = is_bool($bool) ? $bool : false;
                $CC['error'] = false;

                // handle incorrect syntax or incomplete parsed assignments
                // (means: assignment not boolean or placeholders are still
                // inside the assignment)
				if (!is_bool($bool) || preg_match('/\{\$([^#\}]+)\}/', trim($matches[1]))) {
                    // try to parse the assignment separately
					$return = trim($this->fillTemplate($matches[1], false, null, true));

					// if assignments not contains placeholders or the autoRender mode
					// is off, then ervaluate the parsed assignment again,
					// otherwise set handle-flag 'error' to true and the assignment
					// to false (default)
					if (!preg_match('/\{\$([^#\}]+)\}/', $return) || $this->autoRender) {
						// $bool = @call_user_func( @create_function('', 'return (' . $return . ');') );
						$bool = $this->emulate_call_func(trim($return));

	                    $CC['valid'] = is_bool($bool) ? $bool : false;
	                    $CC['error'] = is_bool($bool) ? false : true;
	                } else {
	                	$CC['valid'] = false;
						$CC['error'] = true;
	                }

                    // if an error occured while parsing the assignment and autoRender mode is
                    // on, throw an user error and take out the if-statement
					if ($CC['error'] && $this->autoRender) {
                        trigger_error('The template conditional statement "<em>'.$matches[0].'</em>" at char #' . $_C1 . ' is incorrect (not boolean or syntax error)'
							. ($file? '; file: ' . $template: '') );

						if ( preg_match('/\{\$([^#\}]+)\}/', $matches[0], $matches_) ) {
                        	$matches_ = array_splice($matches_, 1);
							$matches_ = implode(' and ', $matches_);
                        	trigger_error('Reason: unable to parse the template placeholder <em>'.$matches_.'</em>');
                        }
                    }

                    // if an error occured while parsing the assignment and
                    // autoRender mode is off, mark the current statement as
                    // read to denie parsing it again
					if ($CC['error'] && !$this->autoRender) {
                    	$_ = substr($_, 0, $_C1) .
							 substr($matches[0], 0, 1) . strtoupper(uniqid('@CACHED', true)) . substr($matches[0], 1, strlen($matches[0])) .
                     		 substr($_, $_C1, strlen($_) - $_C1);
                    }
                }

                // save the beginning of the current statement into the
                // corresponding handle tag ('start')
				$CC['start'] = $_C1;
            }

            // replace all marked statements with their real identity
			$_ = preg_replace('/\{@CACHED[^\$]+\$/', '{$', $_);

			// -----------------------------------------------------------------

            // -----------------------------------------------------------------
			// php code replacements
            //
            // syntax: {$#PHP statement #}
            //

            while ( preg_match('/\{\$#PHP[[:space:]]([^#]+)#\}/', $_, $matches) ) {
                // try to evaluate php code
				$return = @trim($this->fillTemplate($matches[1], false, null, true));

                // when autoRender is on, replace PHP-Code anyway, otherwise
                // chache this php statement to be matched later again
				if (!preg_match('/\{\$([^#\}]+)\}/', $return) || $this->autoRender) {
					// $return = @call_user_func( @create_function('', 'return (' . $return . ');') );
					$return = $this->emulate_call_func(trim($return));
	                $_ = preg_replace('/\{\$\#PHP[[:space:]]([^#]+)\#\}/', $return, $_, 1);
                } else {
					$_ = preg_replace('/\{\$\#PHP[[:space:]]([^#]+)\#\}/', '{' . strtoupper(uniqid('@CACHED', true)) . '$#PHP \1 #}', $_, 1);
                }

                // handle a DoS (denial of service)
                $x = !isset($x)? 1: $x + 1;
                if ($x>999)
					break;
            }

            $_ = preg_replace('/\{@CACHED[^\$]+\$/si', '{$', $_);

			// -----------------------------------------------------------------

	        // -----------------------------------------------------------------
			// user variable declarations
            //
            // syntax: {$#SET statement #}
            // syntax: {$#GET statement #}
            //
			if ($this->autoRender) {
	            while ( preg_match('/\{\$#SET[[:space:]]([^#]+)#\}/', $_, $matches) ) {
	                // try to evaluate php code
					$return = @trim($this->fillTemplate(preg_replace('/\{\$(\w+)\}/', '$this->pVarStack[\'\1\']', $matches[1]), false, null, true));

					// replace pseuso code TO
					$return = str_replace(' TO ', ' = ', $return);

					// evaluate expression
					eval("{$return};");

					// skip SET statement by empty
					$_ = preg_replace('/\{\$\#SET[[:space:]]([^#]+)\#\}/', '', $_, 1);

	                // handle a DoS (denial of service)
	                $y = !isset($y)? 1: $y + 1;
	                if ($y>999)
						break;
	            }

	            if (isset($y) && $y > 999)
					die('<b>DoS in TTemplate::fillTemplate()</b>');

				unset($y);

	            while ( preg_match('/\{\$#GET[[:space:]]([^#]+)#\}/', $_, $matches) ) {
	            	// try to evaluate php code
					$return = @trim($this->fillTemplate(preg_replace('/\{\$(\w+)\}/', '$this->pVarStack[\'\1\']', $matches[1]), false, null, true));

					// evaluate expression
					eval("\$return = {$return};");

					// replace GET statement by value
					$_ = preg_replace('/\{\$\#GET[[:space:]]([^#]+)\#\}/', $return, $_, 1);

	                // handle a DoS (denial of service)
	                $y = !isset($y)? 1: $y + 1;
	                if ($y>999)
						break;
	            }
	        }
			// -----------------------------------------------------------------

			// handle a DoS (denial of service)

				if (isset($x) && $x > 999)
					die('<b>DoS in TTemplate::fillTemplate()</b>');
				if (isset($y) && $y > 999)
					die('<b>DoS in TTemplate::fillTemplate()</b>');

            // unescape template variable syntax
                // $this->_unescapeTemplateSyntax($this->rows);
                // $this->_unescapeTemplateSyntax($subRows);
                $this->_unescapeTemplateSyntax($_, true);

            return $_;
        }


        /**
         * TTemplate::render()
         *
         * This method does the same line TTemplate::fillTemplate() but
         * enables autoRender before parsing.
         *
         * @param mixed $template
         * @param boolean $file
         * @param array $subRows
         * @return string
         */
        function render($template = null, $file = true, $subRows = null) {
			$_ = $this->autoRender;
			$this->autoRenderOn();
			$string = $this->fillTemplate($template, $file, $subRows);
			$this->autoRender = $_;
			return $string;
        }


        /**
         * TTemplate::destroy_TTemplate()
         *
         * The destructor.
         *
         * @return
         */
        function destroy_TTemplate() {
            // destructor //
        }
    }
?>